from typing import (
    TYPE_CHECKING,
    Awaitable,
    TypeVar,
    Generic,
    NoReturn,
    Callable,
    AsyncIterator,
    Annotated,
    Hashable,
    Sequence,
)
from collections import defaultdict
from collections.abc import Iterable
import itertools
import functools
import asyncio

from .misc import Sentinel

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar("T")
NoneT = TypeVar("NoneT")
DefaultT = TypeVar("DefaultT")
ReturnT = TypeVar("ReturnT")
CallableT = TypeVar("CallableT", bound=Callable)
KeyT = TypeVar("KeyT", bound=Hashable)


def autogather(
    *coros: Awaitable[T] | T | Iterable[Awaitable[T]] | Iterable[T],
    return_exceptions: bool = False
) -> Awaitable[tuple[T | BaseException, ...]]:
    actual_coros = list()
    for i in coros:
        if isinstance(i, Iterable):
            actual_coros.extend(i)
        else:
            actual_coros.append(i)
    return asyncio.gather(
        *map(_autogather_executor, actual_coros),
        return_exceptions=return_exceptions
    )


def _autogather_executor(coro: Awaitable[T] | T) -> T:
    if asyncio.iscoroutine(coro):
        return coro
    return _AwaitableWrap(coro)


class _AwaitableWrap(Generic[T]):
    val: T

    def __init__(self, val: T) -> None:
        self.val = val

    def __await__(self) -> "Self":
        return self

    def __next__(self) -> NoReturn:
        raise StopIteration(self.val)


_mapper_sentinel = Sentinel()


async def amapdefault(
    func: Callable[[T], Awaitable[ReturnT]] | None,
    *iterables: Iterable[T] | NoneT,
    default: DefaultT | None = None,
    default_factory: Callable[[], DefaultT | Awaitable[DefaultT]] | None = None,
    empty_check: bool = False,
    none: NoneT = None,
    check_values_before: bool = False,
    check_values_after: bool = False,
    weak_value_check: bool = False,
    as_list: bool = False
) -> list[ReturnT] | AsyncIterator[ReturnT] | DefaultT:
    """ async version of mapdefault with coro support """

    items = itertools.chain.from_iterable(
        i
        for i in iterables
        if i != none and (not empty_check or i)
    )
    if func is not None:
        items = (
            _mapper_sentinel if (
                check_values_before
                and (
                    (weak_value_check and not i)
                    or i == none
                )
            ) else await func(i)
            for i in items
        )

    result = _amapdefault_generator(
        items,
        none,
        check_values_after,
        weak_value_check
    )
    if not next(result):
        if default_factory is not None:
            default = default_factory()
        if asyncio.iscoroutine(default_factory):
            default = await default
        return default
    if as_list:
        result = [i async for i in result]
    return result


async def _amapdefault_generator(
    items: "Iterable[ReturnT | NoneT | Sentinel]",
    none: NoneT = None,
    check_values_after: bool = False,
    weak_value_check: bool = False
) -> Annotated[
    AsyncIterator[ReturnT | bool],
    bool, ReturnT, ReturnT, ...
]:
    first = True

    for i in items:
        if i is _mapper_sentinel:
            continue
        if check_values_after:
            if not (weak_value_check and i):
                continue
            elif i == none:
                continue

        if first:
            yield True
            first = False
        yield i

    if first:
        yield False


class AsyncEventManager(Generic[KeyT]):
    handlers: dict[KeyT, list[tuple[Callable, bool]]]
    _loop: asyncio.AbstractEventLoop | None

    def __init__(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        self._loop = loop
        self.handlers = defaultdict(list)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    async def emit(self, name: KeyT, *args, **kwargs) -> Sequence:
        return await asyncio.gather(
            *(
                func(name, *args, **kwargs) if with_name else func(*args, **kwargs)
                for func, with_name in self.handlers[name]
            )
        )

    def dispatch(self, name: KeyT, *args, **kwargs) -> asyncio.Task[Sequence]:
        return self.loop.create_task(self.emit(name, *args, **kwargs))

    def dispatcher(self, name: KeyT, *args, **kwargs) -> functools.partial["dispatch"]:
        return functools.partial(self.dispatch, name, *args, **kwargs)

    def on(self, name: KeyT, with_name: bool = False) -> Callable[[CallableT], CallableT]:
        def wrapper(func: CallableT) -> CallableT:
            self.handlers[name].append((func, with_name))
            return func
        return wrapper



from typing import TYPE_CHECKING, Literal
import asyncio
import inspect
import traceback
import types
import logging

if TYPE_CHECKING:
    from typing_extensions import Self

logger = logging.getLogger("yamt")


class ContextSkip(Exception):
    ...


class Grab:
    """ Usage example:
        ```
        async with Grab() as grab, grab.skip:
            ...
        ```
    """

    class Skip:
        grab: "Grab"

        def __init__(self, grab: "Grab") -> None:
            self.grab = grab

        async def __aenter__(self):
            if self.grab.grab_count > 1:
                raise ContextSkip()

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc: BaseException | None,
            tb: types.TracebackType | None
        ):
            ...

    grab_count: int = 0
    skip: Skip

    def __init__(self) -> None:
        self.skip = self.Skip(self)

    async def __aenter__(self) -> "Self":
        self.grab_count += 1
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None
    ) -> Literal[True] | None:
        if isinstance(exc, ContextSkip):
            return True
        self.grab_count -= 1
        return None


# TODO: fix stack inspection in recursive call (rework locked_stack)
class StackLimitedLock(asyncio.Lock):
    stack_limit: int | None
    locked_stack: list[int]

    def __init__(self, stack_limit: int | None = None) -> None:
        logger.warning("unstable synchronization primitive used")
        self.stack_limit = stack_limit
        self.locked_stack = list()
        super().__init__()

    async def acquire(self):
        cur_frame = inspect.currentframe().f_back
        cur_hash = hash(cur_frame.f_code)

        if self.locked():
            for i, (frame, _) in enumerate(traceback.walk_stack(cur_frame)):
                if i == self.stack_limit:
                    break
                if hash(frame.f_code) in self.locked_stack:
                    self.locked_stack.append(cur_hash)
                    return None

        await super().acquire()
        self.locked_stack.append(cur_hash)

    def release(self):
        cur_frame = inspect.currentframe().f_back
        self.locked_stack.remove(hash(cur_frame.f_code))
        if self.locked():
            super().release()


class PerSecondSemaphore(asyncio.Semaphore):
    deffer_time: float
    _lock: asyncio.Lock

    def __init__(self, value: int = 1) -> None:
        super().__init__(value)
        self.deffer_time = 1 / value
        self._lock = asyncio.Lock()

    def release(self):
        asyncio.create_task(self._deffered_exit())

    async def _deffered_exit(self):
        async with self._lock:
            await asyncio.sleep(self.deffer_time)
            super().release()


# deprecated
SemaphorePerSecond = PerSecondSemaphore


class LockOverflowError(Exception):
    ...


class OverflowLock(asyncio.Lock):
    limit: int
    counter: int = 0

    def __init__(self, limit: int) -> None:
        self.limit = limit
        super().__init__()

    async def acquire(self):
        if self.limit <= self.counter:
            raise LockOverflowError()

        await super().acquire()
        self.counter += 1

    def release(self):
        super().release()
        self.counter -= 1


class SkippedOverflowLock(OverflowLock):
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None
    ) -> bool | None:
        if isinstance(exc, LockOverflowError):
            return True
        return await super().__aexit__(exc_type, exc, tb)



from typing import Protocol, Generic, TypeVar, Any

T = TypeVar("T")


class SupportsRichComparison(Protocol):
    def __lt__(self, other: Any) -> bool:
        ...

    def __gt__(self, other: Any) -> bool:
        ...


class SameAs(Generic[T]):
    """ Annotation, marks exactly the same type in context.
        Exists for compatability reasons.
    """
    ...



from typing import TYPE_CHECKING, Any
from enum import EnumMeta, Enum

from .misc import recursive_base_attributes

if TYPE_CHECKING:
    from typing_extensions import Self


# TODO: try to fix typehints (maybe with typestubs?)
class DataRichEnumMeta(EnumMeta):
    @classmethod
    def __prepare__(
        metacls,
        cls: str,
        bases: tuple[type, ...],
        *,
        extender: type | None = None
    ) -> Any:
        return super().__prepare__(cls, bases)

    def __new__(
        metacls,
        cls_name: str,
        bases: tuple[type, ...],
        classdict: dict[str, Any],
        *,
        extender: type | None = None
    ) -> "Self":
        containers: list[tuple[str, Any]] = list()
        containers_names: tuple[str, ...] | None

        if extender is not None:
            annotations: dict | None = classdict.get("__annotations__")
            if annotations is not None:
                annotations.update(extender.__annotations__)
            for k, v in recursive_base_attributes(extender):
                classdict.setdefault(k, v)

        containers_names = classdict.pop("__containers__", None)
        if containers_names is not None:
            assert isinstance(containers_names, tuple)
            for name in containers_names:
                containers.append((name, classdict.pop(name)))

            classdict._member_names = sorted(
                set(classdict._member_names).difference(containers_names),
                key=tuple(classdict._member_names).index
            )
            classdict._ignore.extend(containers_names)
            classdict["_ignore_"] = classdict._ignore

        cls = super().__new__(metacls, cls_name, bases, classdict)

        if containers:
            for k, v in containers:
                setattr(cls, k, v)
        return cls


class DataRichEnum(Enum, metaclass=DataRichEnumMeta):
    ...


class IntDataRichEnum(int, DataRichEnum):
    ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yamt']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'yamt',
    'version': '0.1.6',
    'description': 'yet another monkey toolkit',
    'long_description': '# yamt (yet another monkey toolkit)\n[![PyPI version](https://badge.fury.io/py/yamt.svg)](https://badge.fury.io/py/yamt)\n![PyPI downloads per mounth](https://img.shields.io/pypi/dm/yamt)\n![GitHub code size in bytes](https://img.shields.io/github/languages/code-size/UT1C/yamt)\n\n## Installation\n```\npip install yamt\n```\n',
    'author': 'lightmanLP',
    'author_email': 'liteman1000@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .collections import (
    ChainedSequence,
    ChainedList,
    ChainedDeque,
)
from .enum import (
    DataRichEnum,
    IntDataRichEnum,
)
from .asyncio_sync_primitives import (
    Grab,
    StackLimitedLock,
    SemaphorePerSecond,  # deprecated
    PerSecondSemaphore,
    OverflowLock,
    SkippedOverflowLock,
)
from .misc import (
    WordForm,
    FormedWord,
    Sentinel,
    SingletonMeta,
    IterativeRandomizer,
    DependencyInjector as DI,
    recursive_base_attributes,
    split_on_chuncks,
    simple_chain,
    mapdefault,
    sentinel,
    anyvalue,
)
from .asyncio_misc import (
    AsyncEventManager,
    autogather,
    amapdefault,
)
from .typing import (
    SupportsRichComparison,
)
